import os
import re
import yaml
import json

# Every file matching the pattern gets thrown into a Page object
class Page():
    def __init__(self, fileName, input_dir, root):
        self.file_name, self.fileExtension = os.path.splitext(fileName)
        self.file_path = os.path.join(root, fileName)
        self.relative_path = os.path.join(root.replace(input_dir, "", 1), fileName)
        self.output_filename = ""
        self.output_path = ""
        # self.file_contents = ""
        # self.mtime = 0
        # self.meta = {}
        self.file_contents = self.get_contents(self.file_path)
        self.mtime = self.get_modified_time()
        self.meta = self.get_metadata(self.file_contents)

    def get_contents(self, file_path):
        source_file = open(file_path, 'r')
        file_contents = source_file.read()
        return file_contents

    def get_metadata(self, file_contents):
        page_vars = {}
        if self.fileExtension == ".json":
            page_vars = self.get_as_json(file_contents)

        if self.fileExtension == ".md":
            page_vars = self.get_as_yaml(file_contents)

        return page_vars

    def get_as_json(self, file_contents):

        page_vars = {"yk_data":{"title": None, "template": None, "description": None}}
        default_content = {"data":""}

        json_data = json.loads(file_contents)

        if "yk_data" in json_data:
            page_vars.update(json_data["yk_data"])

        # For backwards compatibility with old stuff
        page_vars["extends"] = page_vars["template"]
        
        if "data" in json_data:
            self.file_contents = json_data["data"]
        else:
            self.file_contents = default_content

        return page_vars

    def get_as_yaml(self, file_contents):
        delimiter = "@@@"
        md_hash = re.split("(%s)" % (delimiter), file_contents)
        page_vars = 0

        if len(md_hash) > 1:
            page_vars = yaml.load(md_hash[2])
            file_contents = "%s%s" % (md_hash[0], md_hash[4])
            self.file_contents = file_contents.replace("{##}", "")

        return page_vars

    def get_page_content(self):
        return self.file_contents

    def get_page_content_from_list(self):
        content = self.file_contents
        values = ','.join(str(v) for v in content)
        return values

    def set_page_content(self, str):
        self.file_contents = str

    def get_page_vars(self):
        return self.meta

    def get_modified_time(self):
        mtime = os.path.getmtime(self.file_path)
        return mtime

    def get_template(self):
        template = 0
        if self.meta != 0:
            if self.meta['extends']:
                template = self.meta['extends']
        return template

    # def before_render(self):
        # self.file_contents = self.get_contents(self.file_path)
        # self.mtime = self.get_modified_time()
        # self.meta = self.get_metadata(self.file_contents)

    def set_output_path(self, output_dir, output_file_extension):
        name = self.relative_path.replace(self.fileExtension, ".%s" % (output_file_extension), 1)
        path = "%s/%s" % (output_dir, name)
        self.output_path = re.sub("/{2,}", "/", path)
        self.output_filename = name

    def get_output_path(self):
        return self.output_path

    def get_file_path(self):
        return self.file_path

    def get_relative_path(self):
        return self.relative_path



import os
import sys
import argparse
import json
from JinjaStaticRenderer import JinjaStaticRenderer

defaults = {
    "generator_config": {
        "build_cache_file": "var/build.cache",
        "default_template": "base.j2",
        "templates_dir": "templates",
        "compiled_templates_file": "var/jinja",
        "compiled_templates_log": "var/jinja.log",
        "output_file_extension": "html",
        "output_dir": "www",
        "input_dir": "site",
        "extensions": ["json"],
        "ignore": [".", "_"],
        "site_vars":"env",
        "verbose": "false"
    },
    "env": {
        "paths": {
            "css": "assets/css",
            "js": "assets/js",
            "assets": "assets",
            "images": "assets/images",
            "webroot": "www",
            "app": "app",
            "content": "content"
        }
    }
}


# Default config just loads the YAML config file
config = {
    'config_path':'config.json'
}

# Generic little app bootstrapper thingy
class App():
    def __init__(self, config):
        self.config = config

    def parse_args(self, args=None):
        d = '\033[32mProcess recursive directories of JSON files through Jinja2 templates. \033[0m'
        parser = argparse.ArgumentParser(description=d)
        parser.add_argument('input_dir',
                            help='\033[34mInput directory such as `site/content`\033[0m', default=defaults['generator_config']['input_dir'])
        parser.add_argument('output_dir',
                            help='\033[34mOutput directory such as `www`\033[0m', default=defaults['generator_config']['output_dir'])
        parser.add_argument("-t", "--templates", action="store_true",
                            help='\033[34mTemplate directory such as `templates`\033[0m', default=defaults['generator_config']['templates_dir'])
        parser.add_argument("-c", "--config", action="store_true",
                            help='\033[34mPath to `config.json`\033[0m', default=config['config_path'])
        parser.add_argument("-o", "--output_file_extension", action="store_true",
                            help='\033[34mType of files to output such as `html`\033[0m', default=defaults['generator_config']['output_file_extension'])
        parser.add_argument("-i", "--input_file_extension", action="store_true",
                            help='\033[34mType of files to match in input_dir, such as `json`\033[0m', default=defaults['generator_config']['extensions'])
        parser.add_argument("-v", "--verbose", action="store_true",
                            help='\033[34mPrint all the plumbing\033[0m')

        return parser.parse_args(args)

    def main(self, args=None):
        args = self.parse_args(args)

        # Pull in extra config from user file
        if os.path.exists( config['config_path'] ):
            userConfig = json.load( file(config['config_path']) )
        else:
            userConfig = defaults

        config.update( userConfig["generator_config"] )
        
        self.config.update(args.__dict__)
        
        # The site vars object is mapped to an item in the json object
        config["site_vars"] = userConfig[userConfig["generator_config"]["site_vars"]]

        renderer = JinjaStaticRenderer(self.config)
        renderer.go()


# Do the thing! 
def start():
    sys.exit(App(config).main())

if __name__ == '__main__':    
    start()




"""
Jinja2 extensions for use Markdown in templates.
"""

import jinja2
import jinja2.ext
import markdown


class MarkdownExtension(jinja2.ext.Extension):
    """
    Block tag for direct code generation from markdown language.

    Original: http://www.silassewell.com/blog/2010/05/10/jinja2-markdown-extension/
    Author: Silas Sewell
    """

    tags = set(['markdown'])

    def __init__(self, environment):
        """
        Jinja2 environvent initialozation.

        :param environment: - page environment;
        :return:            - None.
        """
        super(MarkdownExtension, self).__init__(environment)

    def parse(self, parser):
        """
        Parse template code.

        :param parser:  - Jinja2 parser;
        :return:        - markdown result.
        """
        return False

        if hasattr(self.environment, 'mkdargs'):
            self.environment.extend(
                markdowner=markdown.Markdown(**self.environment.mkdargs))
        else:
            self.environment.extend(
                markdowner=markdown.Markdown())

        lineno = parser.stream.next().lineno
        body = parser.parse_statements(
            ['name:endmarkdown'],
            drop_needle=True
        )
        return jinja2.nodes.CallBlock(
            self.call_method('_markdown_support'),
            [],
            [],
            body
        ).set_lineno(lineno)

    def _markdown_support(self, caller):
        """
        Parse template with markdown.

        :param caller:  - caller of method;
        :return:        - parsed template.
        """
        return self.environment.markdowner.convert(caller()).strip()


from markdown.extensions import Extension
from markdown import inlinepatterns

from markdown.blockprocessors import BlockProcessor, ListIndentProcessor
from markdown.util import etree

DEL_RE = r'(--)(.*?)--'


class MarkdownCustom(BlockProcessor):

    def test(self, parent, block):
        return bool(self.RE.search(block))

    # def run(self, parent, blocks):
        

class MarkdownCustomBlock(Extension):
    def extendMarkdown(self, md, md_globals):
        """ Add an instance of DefListProcessor to BlockParser. """
        md.parser.blockprocessors.add('deflist', 
                                      MarkdownCustom(md.parser),
                                      '>ulist')


# The markdown library expects this to be here to register the ext.
def makeExtension(configs=None):
    return MarkdownCustomBlock(configs=configs)


import os

# Takes a directory and transforms the matching elements into the matching class.
class DirectoryWalker():
    def __init__(self, params):
        self.item_class = params['item_class']
        self.input_dir = params['input_dir']
        self.ignore_files = params['ignore_files']
        self.valid_extensions = params['valid_extensions']

    def fetch(self):
        items = []
        # Walk the input directory and ignore hidden files, partials
        for root, dirs, files in os.walk(self.input_dir):
            for file in files:
                # Ignore hidden files etc
                if any(file.startswith(x) for x in self.ignore_files):
                   continue;
                # Process valid extensions
                if any(file.endswith(x) for x in self.valid_extensions):
                    item = self.item_class(file, self.input_dir, root)
                    items.append(item)
        return items


import os
import time
import markdown as md
import sys
import traceback

from docutils.core import publish_parts
from Page import Page
from DirectoryWalker import DirectoryWalker
from jinja2 import Environment, FileSystemLoader, meta, BaseLoader, TemplateNotFound, ChoiceLoader, Template, Markup
from SilentUndefined import SilentUndefined
from MarkdownExtension import MarkdownExtension


def rst_filter(s):
    return Markup(publish_parts(source=s, writer_name='html')['body'])

def markdown_filter(value):
    """
    Jinja2 markdown filter with support for extensions.
    """
    output = value
    marked = md.Markdown(
        extensions=['extra','meta','sane_lists'],
        output_format='html5',
        safe_mode=False 
        )
    out = marked.convert(output)
    return out

class JinjaStaticRenderer():
    """
    In theory this should probably sub-class a generic renderer
    """ 
    def __init__(self, config):
        self.config = config
        # Setup the env
        self.env = Environment(
            undefined=SilentUndefined,
            loader=FileSystemLoader(config['templates_dir']),
            extensions=[
                MarkdownExtension
            ],
            trim_blocks = True,
            lstrip_blocks = True
            )
        
        self.env.filters["markdown"] = markdown_filter
        self.env.filters['rst'] = rst_filter        
        self.template_modified_times = {}
        self.template_trees = {}

        # self.template_modified_times = self.get_template_mtimes(config)
        # self.template_trees = self.load_abstract_syntax_trees()

        var_path=os.path.dirname(self.config['compiled_templates_file'])


        if self.config["verbose"] == True:
            print "Ensure directory exists: \033[34m%s\033[0m" % (var_path)
            print "Ensure directory exists: \033[34m%s\033[0m" % (config['output_dir'])

        self.ensure_dir(var_path)
        self.ensure_dir(config['output_dir'])

        self.env.compile_templates(
            self.config['compiled_templates_file'],
            ignore_errors=False
            )

        if self.config["verbose"] == True:
            print "Loading templates from: \033[34m%s\033[0m" % (config['templates_dir'])
            print "Compile templates to .zip: \033[32m%s\033[0m" % (self.config['compiled_templates_file'])

        # This lets us have bespoke template pages in the input dir too. 
        # self.env.loader = FileSystemLoader(self.config['templates_dir'])

    # def get_template_mtimes(self, config):
    #     """
    #     Get the template modified times to save disk lookups for each item.
    #     """
    #     template_files = self.env.list_templates()

    #     mod_times = {}

    #     for path in template_files:
    #         file_name = path
    #         file_mtime = os.path.getmtime("%s/%s" % (self.config['templates_dir'], file_name)  )
    #         mod_times[file_name] = file_mtime

    #     return mod_times

    # def load_abstract_syntax_trees(self):
    #     """
    #     Cache up the ASTs so we don't have to hit the disk for each page.
    #     So far, this is probably the single biggest speed boost.
    #     """
    #     template_files = self.env.list_templates()
    #     asts = {}
    #     for template in template_files:
    #         asts[template] = self.env.parse(self.env.loader.get_source(self.env, template))

    #     return asts

    def get_template_mtime(self, template):
        """
        Request the template's mtime. If we don't have it yet, cache it.
        """
        if not template in self.template_modified_times:
            self.template_modified_times[template] = os.path.getmtime("%s/%s" % (self.config['templates_dir'], template) )

        return self.template_modified_times[template]

    def get_referenced_templates(self, template):
        """
        Request the template's Abstract Syntax Tree so we can find other template references.
        Store the reference list in a dictionary
        """
        if not template in self.template_trees:

            try: 
                ast = self.env.parse(self.env.loader.get_source(self.env, template))
                self.template_trees[template] = list(meta.find_referenced_templates(ast))
                
            except:
                self.template_trees[template] = list()

            if self.config["verbose"] == True:
                print "\033[34m%s\033[0m references templates \033[32m%s\033[0m" % (template, self.template_trees[template])

        return self.template_trees[template]

    def get_global_vars(self):
        return self.config['site_vars']

    def ensure_dir(self, dir):
        try:
            os.stat(dir)
        except:
            os.mkdir(dir)

    def render(self, data_source):
        """
        Kick off the rendering loop. If no cache found, or config has been modified, 
        force everything to be re-rendered.
        """
        force_render = 0
        last_build_time = self.get_last_build_time()
        config_modified_time = self.get_config_modified_time()
        output_dir = self.config['output_dir']
        output_file_extension = self.config['output_file_extension']
        templates_dir = self.config['templates_dir']
        default_template = self.config['default_template']
        
        if last_build_time == 0 or config_modified_time >= last_build_time:
            force_render = 1

        rendered_files = []
        
        for item in data_source:
            item.set_output_path(output_dir, output_file_extension)
            rendered_file = self.render_item(last_build_time, self.env, force_render, templates_dir, default_template, item)
            rendered_files.append(rendered_file)



        self.set_last_build_time(time.time())
        self.update_log(rendered_files, last_build_time)


    def render_item(self, last_build_time, env, force_render, templates_dir, default_template, page):
        new_file_path = page.get_output_path()
        new_directory = os.path.split(new_file_path)[0]
        
        # Force a render if the output file doesn't exist
        if not os.path.exists(new_file_path) or page.get_modified_time() >= last_build_time:
            force_render = 1

        template_modified_time = 0
        template = page.get_template()
        
        if template == 0:
            template = default_template

        try:
            template_object = env.get_template(template)
        except:
            print "\033[1;91mCouldn't parse `%s`. Check the template for errors \033[0;37m" % (template)
            
            if self.config["verbose"] == True:
                traceback.print_exc()
            print "\033[0m"
            return [new_file_path, 0, template]

        referenced_templates = self.get_referenced_templates(template)
        template_modified_time = self.get_template_mtime(template)

        if template_modified_time >= last_build_time:
            force_render = 1

        ref_template_mtime = 0
        for t in referenced_templates:
            time = self.get_template_mtime(t)
            if time >= last_build_time:
                force_render = 1
                ref_template_mtime = time

        if self.config["verbose"] == True:
            print "\033[34m%s\033[37m last built at: %s, page mtime: %s, templ mtime: %s, ref-templates mtime: %s\033[0m" % (page.file_path, last_build_time, page.get_modified_time(), template_modified_time, ref_template_mtime)


        file_did_render = 0

        if template_object and last_build_time == 0 or force_render == 1:
            self.ensure_dir(new_directory)
            html = template_object.render(
                content=page.get_page_content(),
                site=self.get_global_vars(),
                template=page.get_page_vars()
                )
            new_file = open(new_file_path, "w")
            new_file.write(html)
            self.print_stdout(page.get_file_path(), new_file_path, template)
            file_did_render = 1

        return [new_file_path, file_did_render, template]
            
    def print_stdout(self, original_path, new_path, template):
        print "\033[32mBuilding \033[0m\033[2m%s\033[0m > \033[34m%s \033[2m[%s]\033[0m" % (original_path, new_path, template)

    def update_log(self, items, last_modified_time):
        file = open(self.config['compiled_templates_log'], 'w')
        print>>file, "this_render=%s, prev_render=%s" % (time.time(), last_modified_time)

        for item in items:
            print>>file, "%s, %s, %s" % (item[0], item[2], item[1])

        if self.config["verbose"] == True:
            print "Wrote log to: \033[32m%s\033[0m" % (self.config['compiled_templates_log'])

    def get_last_build_time(self):
        """
        Get the last run time. Useful for checking what stuff's changed since we
        last ran the script.
        """
        last_time = 0
        if os.path.exists(self.config['build_cache_file']):
            with open(self.config['build_cache_file'], "r") as buildFile: 
                lines = map(float, buildFile.readline().split())
                if len(lines) > 0:
                    last_time = lines[0]
        return last_time

    def get_config_modified_time(self):
        """
        Force a re-render if the config file has changed (global paths may be invalid, etc..)
        """
        mtime = 0
        if os.path.exists(self.config['config_path']):
            mtime = os.path.getmtime(self.config['config_path'])
        return mtime

    def set_last_build_time(self, update_time):
        """
        Save the last successful run time in a var
        """
        with open(self.config['build_cache_file'], "w") as buildFile:
            buildFile.write("%s" % (update_time))

    def get_data(self, source_type, params):
        """
        Future-friendly method. For when this interfaces with a swanky API.
        """
        if source_type == "directory":
            source = DirectoryWalker(params)
            return source.fetch()

    def go(self):
        """
        Request some data from an endpoint/directory
        """
        data = self.get_data(
            "directory",
            {
                'input_dir': self.config['input_dir'],
                'valid_extensions': self.config['extensions'],
                'ignore_files': self.config['ignore'],
                'item_class': Page
            }
        )

        self.render(data)